require 'uri'
require 'simple-hmac'
require 'rest_client'
require 'active_model'
require 'active_model/model'
require 'active_model/serializers/json'
require 'active_support/core_ext/hash/indifferent_access'
require 'active_support/core_ext/object/to_param'
require 'active_support/core_ext/module/remove_method'

module ApiResource

  class ResourceError < StandardError
    attr_reader :model

    def initialize(model, message)
      super(message)
      @model = model
    end
  end

  class Resource
    include ActiveModel::Model

    class_attribute :base_url
    class_attribute :hmac
    class_attribute :api_id
    class_attribute :api_secret
    class_attribute :hmac_options
    class_attribute :pre_submit_filter
    class_attribute :default_id_attr
    self.default_id_attr = :id
    class_attribute :default_per_page
    self.default_per_page = 500
    class_attribute :logger

    attr_reader :raw_result

    def self.with_hmac(api_id, api_secret, options={})
      self.hmac         = true
      self.api_id       = api_id
      self.api_secret   = api_secret
      self.hmac_options = options
    end

    def self.filter_submitted_attributes(&block)
      self.pre_submit_filter = block if block_given?
    end

    def initialize(hash=nil, raw_result=nil)
      self.attributes = hash if hash
      @raw_result = raw_result
    end

    def self.resource_path
      @resource_path || name.to_s.tableize.split('/').last
    end

    def self.resource_path=(path)
      @resource_path = path.gsub(%r{\A/+|/+\Z}, '')
    end

    def resource_path
      @resource_path ||= self.class.resource_path
    end

    def resource_path=(path)
      @resource_path = path
    end

    def self.find(id, params={})
      result = client(:get, params, resource_path, id)
      json   = JSON.parse(result)
      self.new(json['data'] || json, result)
    end

    def self.all
      self.where({})
    end

    def self.all_pages(params={})
      params  = { page: 1, per_page: default_per_page }.merge(params)
      per_page = params[:per_page].to_i
      page     = params[:page].to_i
      begin
        params[:page] = page
        result         = self.where(params)
        total_count    = result.try :total_count
        arr            = result.to_a
        count          = arr.length
        break if yield(arr, page, per_page, total_count) || (page - 1) * per_page + count >= total_count
        page += 1
      end while true
    end

    def self.where(params)
      result = client(:get, params, resource_path)
      create_resource_collection(result)
    end

    def self.destroy!(id)
      client(:delete, {}, resource_path, id)
      self
    end

    def save!(options={})
      submit_resource(resource_path, true, options[:id_attr])
      self
    end

    def submit!(options={})
      path = options.fetch(:path, resource_path)
      type = options[:type]
      json = submit_resource(path, false, options[:id_attr])
      meta = json && json['meta']
      type ||= meta && meta['type']
      if type
        returned_type = self.class.load_class(type)
        returned_type.new(json['data'] || json, raw_result)
      else
        self
      end
    end

    def destroy!(options={})
      submit_resource(resource_path, false, options[:id_attr], :delete)
      self
    end

    def attributes
      instance_values.with_indifferent_access.except(:errors, :validation_context, :resource_path)
    end

    protected

    def parse_and_check_error(result)
      return nil if result.blank?
      json        = JSON.parse(result)
      errors_hash = json['errors']
      if errors_hash.is_a?(Hash)
        errors_hash.each { |attr, ary| ary.each { |error| errors.add(attr, error) } }
        nil
      else
        json
      end
    end

    def attributes=(hash)
      hash.each do |key, value|
        if respond_to? "#{key}="
          send("#{key}=", value)
        else
          # define method for associated embedded resource(s)
          if value.is_a?(Hash)
            meta, data = value['meta'], value['data'] || value
          else
            meta, data = nil, value
          end

          child_class = self.class.load_class((meta && meta['type']) || key)
          child_value = if child_class && data.is_a?(Hash)
                          child_class.new(data)
                        elsif child_class && data.is_a?(Array)
                          ResourceCollection.new(data, meta, child_class)
                        else
                          data
                        end

          self.class.remove_possible_method(key)
          define_singleton_method(key) { child_value }
        end
      end
    end

    def self.method_missing(m, *args, &_)
      case (m)
        when /^by_(.+)/
          my_resource_path = build_path($1.pluralize, args[0], self.resource_path)
          my_model_name = self.model_name
          Class.new(self) do
            self.resource_path = my_resource_path
            @model_name = my_model_name
            def self.model_name
              @model_name
            end
          end
        else
          super
      end
    end

    def method_missing(m, *args, &_)
      case m
        when /^by_(.+)/
          self.resource_path = self.class.build_path($1.pluralize, args[0], resource_path)
          self
        else
          super
      end
    end

    def self.create_resource_collection(result)
      json = JSON.parse(result)
      ResourceCollection.new(json['data'], json['meta'], self, result)
    end

    def self.load_class(tableized_class)
      my_name = name || ancestors.map(&:name).find {|e| e}
      tableized_class = my_name.to_s.tableize.split('/')[0...-1].push(tableized_class).join('/')
      klass = tableized_class && tableized_class.to_s.singularize.classify.constantize rescue nil
      klass if klass && klass < ApiResource::Resource
    end

    def submit_resource(path, save_attrs=false, id_attr=nil, method=nil)
      raise ResourceError.new(self, 'Attempt to submit invalid resource') unless valid? || method==:delete
      id_attr  = id_attr || default_id_attr
      path     = [path]
      attrs    = attributes
      id_value = attrs[id_attr]
      method   = method || (id_value.nil? ? :post : :put)

      if method != :post && id_value
        path << id_value
        attrs.delete(id_attr)
      end
      attrs = {} if method == :delete
      attrs = pre_submit_filter[attrs] if pre_submit_filter
      begin
        result = self.class.client(method, attrs, *path)
      rescue RestClient::ExceptionWithResponse => e
        result = e.http_body
        raise ResourceError.new(self, "Error executing request #{e.http_code} #{e.http_body}")
      ensure
        json            = parse_and_check_error(result)
        self.attributes = json['data'] || json if json && method == :post && save_attrs
      end
      json
    end

    def self.client(method, params, *paths)
      url        = build_url(*paths)
      headers    = { accept: :json }
      req_params = { url: url, method: method, headers: headers }
      if method == :get
        # using #to_query instead of URI::encode_www_form to comply with rails convention for query string array values
        req_params[:url] = "#{url}?#{params.to_query}" unless params.to_query.blank?
      else
        req_params[:payload] = params
      end
      req = RestClient::Request.new(req_params)
      if hmac
        req.sign!(api_id, api_secret, hmac_options)
      end
      result = req.execute
      log << result
      result
    end

    def self.build_url(*paths)
      raise RuntimeError, 'Empty base_url' if base_url.blank?
      URI::parse(base_url).merge(build_path(*paths)).to_s
    end

    def self.build_path(*paths)
      paths.map { |p| URI.encode p.to_s.gsub(%r{\A/*(.*?)/*\z}, '\1') }.join('/')
    end

    def self.log=(param)
      self.logger = create_log(param)
    end

    def self.log
      self.logger ||= Class.new do
        def <<(_)
        end
      end.new
    end

    # Create a log that respond to << like a logger
    # param can be 'stdout', 'stderr', a string (then we will log to that file) or a logger (then we return it)
    def self.create_log(param)
      return param unless param.is_a? String
      case param
        when 'stdout'
          Class.new do
            def <<(obj)
              STDOUT.puts obj
            end
          end.new
        when 'stderr'
          Class.new do
            def <<(obj)
              STDERR.puts obj
            end
          end.new
        else
          Class.new do
            def initialize(file)
              @file = file
            end

            def <<(obj)
              File.open(@file, 'a') { |f| f.puts obj }
            end
          end.new(param)
      end
    end
  end

  class ResourceCollection
    include Enumerable

    attr_reader :raw_result

    def initialize(data, meta, klass, raw_result=nil)
      meta.each { |k, v| self.class.class_eval { define_method(k) { v } } } if meta
      @resources = data.map { |e| klass.new(e) }.to_a
      @raw_result = raw_result
    end

    def method_missing(m, *args, &block)
      if @resources.respond_to?(m)
        @resources.send(m,*args, &block)
      else
        super
      end
    end

    def respond_to_missing?(m, _)
      @resources.respond_to?(m)
    end

    def acts_like_array?
    end
  end
end
